@@ -10,15 +10,14 @@
 #include <stdio.h>
 
 #include "../processx.h"
+#include "../cleancall.h"
 
 /* Internals */
 
-static void processx__child_init(processx_handle_t *handle, int (*pipes)[2],
-				 int stdio_count, char *command, char **args,
-				 int error_fd, const char *std_in,
-				 const char *std_out,
-				 const char *std_err,
-                                 const char *pty_name, char **env,
+static void processx__child_init(processx_handle_t *handle, SEXP connections,
+				 int (*pipes)[2], int stdio_count,
+				 char *command, char **args,
+				 int error_fd, const char *pty_name, char **env,
                                  processx_options_t *options,
 				 const char *tree_id);
 
@@ -42,7 +41,7 @@ void processx__create_connections(processx_handle_t *handle, SEXP private,
 #endif
 #endif
 
-#if defined(__APPLE__) && !TARGET_OS_IPHONE
+#if defined(__APPLE__)
 # include <crt_externs.h>
 # define environ (*_NSGetEnviron())
 #else
@@ -57,6 +56,8 @@ extern processx__child_list_t *child_list;
 extern processx__child_list_t child_free_list_head;
 extern processx__child_list_t *child_free_list;
 
+extern int processx__notify_old_sigchld_handler;
+
 /* We are trying to make sure that the variables in the library are
    properly set to their initial values after a library (re)load.
    This function is called from `R_init_processx`. */
@@ -71,9 +72,13 @@ void R_init_processx_unix() {
   child_free_list_head.weak_status = R_NilValue;
   child_free_list_head.next = 0;
   child_free_list = &child_free_list_head;
+
+  if (getenv("PROCESSX_NOTIFY_OLD_SIGCHLD")) {
+    processx__notify_old_sigchld_handler = 1;
+  }
 }
 
-int processx__pty_master_open(char *slave_name, size_t sn_len) {
+int processx__pty_master_open(char *sub_name, size_t sn_len) {
   int master_fd, saved_errno;
   char *p;
 
@@ -103,7 +108,7 @@ int processx__pty_master_open(char *slave_name, size_t sn_len) {
   }
 
   if (strlen(p) < sn_len) {
-    strncpy(slave_name, p, sn_len);
+    strncpy(sub_name, p, sn_len);
   } else {
     close(master_fd);
     errno = EOVERFLOW;
@@ -121,17 +126,14 @@ void processx__write_int(int fd, int err) {
   (void) dummy;
 }
 
-static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
-				 int stdio_count, char *command, char **args,
-				 int error_fd, const char *std_in,
-				 const char *std_out,
-				 const char *std_err,
-                                 const char *pty_name, char **env,
-				 processx_options_t *options,
-				 const char *tree_id) {
+static void processx__child_init(processx_handle_t *handle, SEXP connections,
+                                 int (*pipes)[2], int stdio_count,
+                                 char *command, char **args,
+                                 int error_fd, const char *pty_name, char **env,
+                                 processx_options_t *options,
+                                 const char *tree_id) {
 
   int close_fd, use_fd, fd, i;
-  const char *out_files[3] = { std_in, std_out, std_err };
   int min_fd = 0;
 
   setsid();
@@ -141,14 +143,24 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
     /* Do not mess with stdin/stdout/stderr, all handled by the pty */
     min_fd = 3;
 
-    int slave_fd = open(pty_name, O_RDWR);
-    if (slave_fd == -1) {
+    int sub_fd = open(pty_name, O_RDWR);
+    if (sub_fd == -1) {
       processx__write_int(error_fd, -errno);
       raise(SIGKILL);
     }
 
 #ifdef TIOCSCTTY
-    if (ioctl(slave_fd, TIOCSCTTY, 0) == -1) {
+    if (ioctl(sub_fd, TIOCSCTTY, 0) == -1) {
+      processx__write_int(error_fd, -errno);
+      raise(SIGKILL);
+    }
+#endif
+
+#ifdef TIOCSWINSZ
+    struct winsize w;
+    w.ws_row = options->pty_rows;
+    w.ws_col = options->pty_cols;
+    if (ioctl(sub_fd, TIOCSWINSZ, &w) == -1) {
       processx__write_int(error_fd, -errno);
       raise(SIGKILL);
     }
@@ -156,7 +168,7 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
 
     struct termios tp;
 
-    if (tcgetattr(slave_fd, &tp) == -1) {
+    if (tcgetattr(sub_fd, &tp) == -1) {
       processx__write_int(error_fd, -errno);
       raise(SIGKILL);
     }
@@ -167,28 +179,28 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
       tp.c_lflag &= ~ECHO;
     }
 
-    if (tcsetattr(slave_fd, TCSAFLUSH, &tp) == -1) {
+    if (tcsetattr(sub_fd, TCSAFLUSH, &tp) == -1) {
       processx__write_int(error_fd, -errno);
       raise(SIGKILL);
     }
 
     /* TODO: set other terminal attributes and size */
 
-    /* Duplicate pty slave to be child's stdin, stdout, and stderr */
-    if (dup2(slave_fd, STDIN_FILENO) != STDIN_FILENO) {
+    /* Duplicate pty sub to be child's stdin, stdout, and stderr */
+    if (dup2(sub_fd, STDIN_FILENO) != STDIN_FILENO) {
       processx__write_int(error_fd, -errno);
       raise(SIGKILL);
     }
-    if (dup2(slave_fd, STDOUT_FILENO) != STDOUT_FILENO) {
+    if (dup2(sub_fd, STDOUT_FILENO) != STDOUT_FILENO) {
       processx__write_int(error_fd, -errno);
       raise(SIGKILL);
     }
-    if (dup2(slave_fd, STDERR_FILENO) != STDERR_FILENO) {
+    if (dup2(sub_fd, STDERR_FILENO) != STDERR_FILENO) {
       processx__write_int(error_fd, -errno);
       raise(SIGKILL);
     }
 
-    if (slave_fd > STDERR_FILENO) close(slave_fd);
+    if (sub_fd > STDERR_FILENO) close(sub_fd);
   }
 
   /* We want to prevent use_fd < fd, because we will dup2() use_fd into
@@ -213,6 +225,10 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
      process properly. */
 
   for (fd = min_fd; fd < stdio_count; fd++) {
+    SEXP output = VECTOR_ELT(connections, fd);
+    const char *stroutput =
+      Rf_isString(output) ? CHAR(STRING_ELT(output, 0)) : NULL;
+
     /* close_fd is an fd that must be closed. Initially this is the
        parent's end of a pipe. (-1 if no pipe for this fd.) */
     close_fd = pipes[fd][0];
@@ -220,8 +236,8 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
     use_fd = pipes[fd][1];
 
     /* If no pipe, then we see if this is the 2>&1 case. */
-    if (fd == 2 && use_fd < 0 && out_files[fd] &&
-	! strcmp(out_files[fd], "2>&1")) {
+    if (fd == 2 && use_fd < 0 && stroutput &&
+	! strcmp(stroutput, "2>&1")) {
       use_fd = 1;
 
     } else if (use_fd < 0) {
@@ -231,12 +247,12 @@ static void processx__child_init(processx_handle_t* handle, int (*pipes)[2],
 	 no need to open any file */
       if (fd >= 3) continue;
 
-      if (out_files[fd]) {
+      if (stroutput) {
 	/* A file was requested, open it */
 	if (fd == 0) {
-	  use_fd = open(out_files[fd], O_RDONLY);
+	  use_fd = open(stroutput, O_RDONLY);
 	} else {
-	  use_fd = open(out_files[fd], O_CREAT | O_TRUNC| O_RDWR, 0644);
+	  use_fd = open(stroutput, O_CREAT | O_TRUNC| O_RDWR, 0644);
 	}
       } else {
 	/* NULL, so stdin/out/err is ignored, using /dev/null */
@@ -356,7 +372,7 @@ static SEXP processx__make_handle(SEXP private, int cleanup) {
   SEXP result;
 
   handle = (processx_handle_t*) malloc(sizeof(processx_handle_t));
-  if (!handle) { R_THROW_ERROR("Out of memory"); }
+  if (!handle) { R_THROW_ERROR("Cannot make processx handle, out of memory"); }
   memset(handle, 0, sizeof(processx_handle_t));
   handle->waitpipe[0] = handle->waitpipe[1] = -1;
 
@@ -373,7 +389,7 @@ static void processx__handle_destroy(processx_handle_t *handle) {
   free(handle);
 }
 
-void processx__make_socketpair(int pipe[2]) {
+void processx__make_socketpair(int pipe[2], const char *exe) {
 #if defined(__linux__)
   static int no_cloexec;
   if (no_cloexec)  goto skip;
@@ -394,31 +410,34 @@ skip:
 #endif
 
   if (socketpair(AF_UNIX, SOCK_STREAM, 0, pipe)) {
-    R_THROW_SYSTEM_ERROR("processx socketpair");
+    if (exe) {
+      R_THROW_SYSTEM_ERROR("cannot make processx socketpair while "
+                           "running '%s'", exe);
+    } else {
+      R_THROW_SYSTEM_ERROR("cannot make processx socketpair");
+    }
   }
 
   processx__cloexec_fcntl(pipe[0], 1);
   processx__cloexec_fcntl(pipe[1], 1);
 }
 
-SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
-		   SEXP std_err, SEXP pty, SEXP pty_options,
+SEXP processx_exec(SEXP command, SEXP args, SEXP pty, SEXP pty_options,
                    SEXP connections, SEXP env, SEXP windows_verbatim_args,
-                   SEXP windows_hide_window, SEXP private, SEXP cleanup,
-                   SEXP wd, SEXP encoding, SEXP tree_id) {
+                   SEXP windows_hide_window, SEXP windows_detached_process,
+                   SEXP private, SEXP cleanup, SEXP wd, SEXP encoding,
+                   SEXP tree_id) {
 
   char *ccommand = processx__tmp_string(command, 0);
   char **cargs = processx__tmp_character(args);
   char **cenv = isNull(env) ? 0 : processx__tmp_character(env);
   int ccleanup = INTEGER(cleanup)[0];
-  const char *cstdin = isNull(std_in) ? 0 : CHAR(STRING_ELT(std_in, 0));
-  const char *cstdout = isNull(std_out) ? 0 : CHAR(STRING_ELT(std_out, 0));
-  const char *cstderr = isNull(std_err) ? 0 : CHAR(STRING_ELT(std_err, 0));
+
   const int cpty = LOGICAL(pty)[0];
   const char *cencoding = CHAR(STRING_ELT(encoding, 0));
   const char *ctree_id = CHAR(STRING_ELT(tree_id, 0));
   processx_options_t options = { 0 };
-  int num_connections = LENGTH(connections) + 3;
+  int num_connections = LENGTH(connections);
 
   pid_t pid;
   int err, exec_errorno = 0, status;
@@ -440,7 +459,7 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
   options.wd = isNull(wd) ? 0 : CHAR(STRING_ELT(wd, 0));
 
   if (pipe(signal_pipe)) {
-    R_THROW_SYSTEM_ERROR("Cannot create pipe");
+    R_THROW_SYSTEM_ERROR("Cannot create pipe when running '%s'", ccommand);
   }
   processx__cloexec_fcntl(signal_pipe[0], 1);
   processx__cloexec_fcntl(signal_pipe[1], 1);
@@ -454,22 +473,47 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
     pty_master_fd =
       processx__pty_master_open(pty_name, R_PROCESSX_PTY_NAME_LEN);
     if (pty_master_fd == -1) {
-      R_THROW_SYSTEM_ERROR("Cannot open pty");
+      R_THROW_SYSTEM_ERROR("Cannot open pty when running '%s'", ccommand);
     }
     options.pty_echo = LOGICAL(VECTOR_ELT(pty_options, 0))[0];
-
-  } else {
-    /* Create pipes, if requested. */
-    if (cstdin && !strcmp(cstdin, "|")) processx__make_socketpair(pipes[0]);
-    if (cstdout && !strcmp(cstdout, "|")) processx__make_socketpair(pipes[1]);
-    if (cstderr && !strcmp(cstderr, "|")) processx__make_socketpair(pipes[2]);
+    options.pty_rows = INTEGER(VECTOR_ELT(pty_options, 1))[0];
+    options.pty_cols = INTEGER(VECTOR_ELT(pty_options, 2))[0];
   }
 
-  for (i = 0; i < num_connections - 3; i++) {
-    processx_connection_t *ccon =
-      R_ExternalPtrAddr(VECTOR_ELT(connections, i));
-    int fd = processx_c_connection_fileno(ccon);
-    pipes[i + 3][1] = fd;
+  handle->fd0 = handle->fd1 = handle->fd2 = -1;
+  for (i = 0; i < num_connections; i++) {
+    SEXP output = VECTOR_ELT(connections, i);
+    const char *stroutput =
+      Rf_isString(output) ? CHAR(STRING_ELT(output, 0)) : NULL;
+
+    if (isNull(output)) {
+      /* Ignored output, nothing to do, handled in the child */
+
+    } else if (stroutput && ! strcmp("|", stroutput)) {
+      /* pipe, need to create */
+      processx__make_socketpair(pipes[i], ccommand);
+      if (i == 0) handle->fd0 = pipes[i][0];
+      if (i == 1) handle->fd1 = pipes[i][0];
+      if (i == 2) handle->fd2 = pipes[i][0];
+      processx__nonblock_fcntl(pipes[i][0], 1);
+
+    } else if (i == 2 && stroutput && ! strcmp("2>&1", stroutput)) {
+      /* redirected stderr, handled in child */
+
+    } else if (stroutput && ! strcmp("", stroutput)) {
+      /* inherited std stream, assume usual numbers */
+      pipes[i][1] = i;
+
+    } else if (stroutput) {
+      /* redirect to file, nothing to do the child will open it */
+
+    } else {
+      /* inherited processx connection, need to duplicate */
+      processx_connection_t *ccon =
+	R_ExternalPtrAddr(VECTOR_ELT(connections, i));
+      int fd = processx_c_connection_fileno(ccon);
+      pipes[i][1] = fd;
+    }
   }
 
   processx__block_sigchld();
@@ -483,17 +527,20 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
     if (signal_pipe[1] >= 0) close(signal_pipe[1]);
     if (cpty) close(pty_master_fd);
     processx__unblock_sigchld();
-    R_THROW_SYSTEM_ERROR_CODE(err, "Cannot fork");
+    R_THROW_SYSTEM_ERROR_CODE(err, "Cannot fork when running '%s'",
+                              ccommand);
   }
 
   /* CHILD */
   if (pid == 0) {
     /* LCOV_EXCL_START */
     if (cpty) close(pty_master_fd);
-    processx__child_init(handle, pipes, num_connections, ccommand, cargs,
-			 signal_pipe[1], cstdin, cstdout, cstderr,
+    processx__unblock_sigchld();
+    processx__child_init(handle, connections, pipes, num_connections,
+			 ccommand, cargs, signal_pipe[1],
                          pty_name, cenv, &options, ctree_id);
-    R_THROW_SYSTEM_ERROR("Cannot start child process");
+    R_THROW_SYSTEM_ERROR("Cannot start child process when running '%s'",
+                         ccommand);
     /* LCOV_EXCL_STOP */
   }
 
@@ -510,7 +557,8 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
     if (signal_pipe[0] >= 0) close(signal_pipe[0]);
     if (signal_pipe[1] >= 0) close(signal_pipe[1]);
     processx__unblock_sigchld();
-    R_THROW_ERROR("Cannot create child process, out of memory");
+    R_THROW_ERROR("Cannot create child process '%s', out of memory",
+                  ccommand);
   }
 
   /* SIGCHLD can arrive now */
@@ -536,29 +584,16 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
 
   } else {
     R_THROW_SYSTEM_ERROR_CODE(-exec_errorno,
-                              "Child process failed to start");
+                              "Child process '%s' failed to start",
+                              ccommand);
   }
 
   if (signal_pipe[0] >= 0) close(signal_pipe[0]);
 
-  /* Set fds for standard I/O */
-  handle->fd0 = handle->fd1 = handle->fd2 = -1;
-  if (pipes[0][0] >= 0) {
-    handle->fd0  = pipes[0][0];
-    processx__nonblock_fcntl(handle->fd0, 1);
-  }
-  if (pipes[1][0] >= 0) {
-    handle->fd1 = pipes[1][0];
-    processx__nonblock_fcntl(handle->fd1, 1);
-  }
-  if (pipes[2][0] >= 0) {
-    handle->fd2 = pipes[2][0];
-    processx__nonblock_fcntl(handle->fd2, 1);
-  }
-
-  /* Closed unused ends of pipes */
+  /* Closed unused ends of std pipes. If there is no parent end, then
+     this is an inherited std{in,out,err} fd, so we should not close it. */
   for (i = 0; i < 3; i++) {
-    if (pipes[i][1] >= 0) close(pipes[i][1]);
+    if (pipes[i][1] >= 0 && pipes[i][0] >= 0) close(pipes[i][1]);
   }
 
   /* Create proper connections */
@@ -570,7 +605,9 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_in, SEXP std_out,
     return result;
   }
 
-  R_THROW_SYSTEM_ERROR_CODE(-exec_errorno, "cannot start processx process");
+  R_THROW_SYSTEM_ERROR_CODE(-exec_errorno,
+                            "cannot start processx process '%s'",
+                            ccommand);
   return R_NilValue;
 }
 
@@ -599,6 +636,14 @@ void processx__collect_exit_status(SEXP status, int retval, int wstat) {
   handle->collected = 1;
 }
 
+static void processx__wait_cleanup(void *ptr) {
+  int *fds = (int*) ptr;
+  if (!fds) return;
+  if (fds[0] >= 0) close(fds[0]);
+  if (fds[1] >= 0) close(fds[1]);
+  free(fds);
+}
+
 /* In general we need to worry about three asynchronous processes here:
  * 1. The main code, i.e. the code in this function.
  * 2. The finalizer, that can be triggered by any R function.
@@ -624,13 +669,19 @@ void processx__collect_exit_status(SEXP status, int retval, int wstat) {
  * 7. We keep polling until the timeout expires or the process finishes.
  */
 
-SEXP processx_wait(SEXP status, SEXP timeout) {
+SEXP processx_wait(SEXP status, SEXP timeout, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   int ctimeout = INTEGER(timeout)[0], timeleft = ctimeout;
   struct pollfd fd;
   int ret = 0;
   pid_t pid;
 
+  int *fds = malloc(sizeof(int) * 2);
+  if (!fds) R_THROW_SYSTEM_ERROR("Allocating memory when waiting");
+  fds[0] = fds[1] = -1;
+  r_call_on_exit(processx__wait_cleanup, fds);
+
   processx__block_sigchld();
 
   if (!handle) {
@@ -653,8 +704,10 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
   /* Setup the self-pipe that we can poll */
   if (pipe(handle->waitpipe)) {
     processx__unblock_sigchld();
-    R_THROW_SYSTEM_ERROR("processx error");
+    R_THROW_SYSTEM_ERROR("processx error when waiting for '%s'", cname);
   }
+  fds[0] = handle->waitpipe[0];
+  fds[1] = handle->waitpipe[1];
   processx__nonblock_fcntl(handle->waitpipe[0], 1);
   processx__nonblock_fcntl(handle->waitpipe[1], 1);
 
@@ -665,6 +718,8 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
 
   processx__unblock_sigchld();
 
+
+
   while (ctimeout < 0 || timeleft > PROCESSX_INTERRUPT_INTERVAL) {
     do {
       ret = poll(&fd, 1, PROCESSX_INTERRUPT_INTERVAL);
@@ -696,12 +751,12 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
   }
 
   if (ret == -1) {
-    R_THROW_SYSTEM_ERROR("processx wait with timeout error");
+    R_THROW_SYSTEM_ERROR("processx wait with timeout error while "
+                         "waiting for '%s'", cname);
   }
 
  cleanup:
-  if (handle->waitpipe[0] >= 0) close(handle->waitpipe[0]);
-  if (handle->waitpipe[1] >= 0) close(handle->waitpipe[1]);
+  /* pipe is closed in the on_exit handler */
   handle->waitpipe[0] = -1;
   handle->waitpipe[1] = -1;
 
@@ -723,8 +778,9 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
  * 6. Otherwise we collect the exit status, and return FALSE.
  */
 
-SEXP processx_is_alive(SEXP status) {
+SEXP processx_is_alive(SEXP status, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   pid_t pid;
   int wstat, wp;
   int ret = 0;
@@ -750,7 +806,7 @@ SEXP processx_is_alive(SEXP status) {
   /* Some other error? */
   if (wp == -1) {
     processx__unblock_sigchld();
-    R_THROW_SYSTEM_ERROR("processx_is_alive");
+    R_THROW_SYSTEM_ERROR("processx_is_alive, process '%s'", cname);
   }
 
   /* If running, return TRUE, otherwise collect exit status, return FALSE */
@@ -769,8 +825,9 @@ SEXP processx_is_alive(SEXP status) {
  * exit status if the process has already finished. See above.
  */
 
-SEXP processx_get_exit_status(SEXP status) {
+SEXP processx_get_exit_status(SEXP status, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   pid_t pid;
   int wstat, wp;
   SEXP result;
@@ -805,7 +862,7 @@ SEXP processx_get_exit_status(SEXP status) {
   /* Some other error? */
   if (wp == -1) {
     processx__unblock_sigchld();
-    R_THROW_SYSTEM_ERROR("processx_get_exit_status");
+    R_THROW_SYSTEM_ERROR("processx_get_exit_status error for '%s'", cname);
   }
 
   /* If running, do nothing otherwise collect */
@@ -833,8 +890,9 @@ SEXP processx_get_exit_status(SEXP status) {
  * the SIGCHLD handler.
  */
 
-SEXP processx_signal(SEXP status, SEXP signal) {
+SEXP processx_signal(SEXP status, SEXP signal, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   pid_t pid;
   int wstat, wp, ret, result;
 
@@ -861,7 +919,7 @@ SEXP processx_signal(SEXP status, SEXP signal) {
     result = 0;
   } else {
     processx__unblock_sigchld();
-    R_THROW_SYSTEM_ERROR("processx_signal");
+    R_THROW_SYSTEM_ERROR("processx_signal for '%s'", cname);
     return R_NilValue;
   }
 
@@ -878,7 +936,7 @@ SEXP processx_signal(SEXP status, SEXP signal) {
 
   if (wp == -1) {
     processx__unblock_sigchld();
-    R_THROW_SYSTEM_ERROR("processx_signal");
+    R_THROW_SYSTEM_ERROR("processx_signal for '%s'", cname);
   }
 
  cleanup:
@@ -886,8 +944,8 @@ SEXP processx_signal(SEXP status, SEXP signal) {
   return ScalarLogical(result);
 }
 
-SEXP processx_interrupt(SEXP status) {
-  return processx_signal(status, ScalarInteger(2));
+SEXP processx_interrupt(SEXP status, SEXP name) {
+  return processx_signal(status, ScalarInteger(2), name);
 }
 
 /* This is a special case of `processx_signal`, and we implement it almost
@@ -901,8 +959,9 @@ SEXP processx_interrupt(SEXP status) {
  * still alive or not.
  */
 
-SEXP processx_kill(SEXP status, SEXP grace) {
+SEXP processx_kill(SEXP status, SEXP grace, SEXP name) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
+  const char *cname = isNull(name) ? "???" : CHAR(STRING_ELT(name, 0));
   pid_t pid;
   int wstat, wp, result = 0;
 
@@ -929,7 +988,7 @@ SEXP processx_kill(SEXP status, SEXP grace) {
   /* Some other error? */
   if (wp == -1) {
     processx__unblock_sigchld();
-    R_THROW_SYSTEM_ERROR("processx_kill");
+    R_THROW_SYSTEM_ERROR("processx_kill for '%s'", cname);
   }
 
   /* If the process is not running, return (FALSE) */
@@ -940,7 +999,7 @@ SEXP processx_kill(SEXP status, SEXP grace) {
   if (ret == -1 && (errno == ESRCH || errno == EPERM)) { goto cleanup; }
   if (ret == -1) {
     processx__unblock_sigchld();
-    R_THROW_SYSTEM_ERROR("process_kill");
+    R_THROW_SYSTEM_ERROR("process_kill for '%s'", cname);
   }
 
   /* Do a waitpid to collect the status and reap the zombie */
@@ -983,7 +1042,7 @@ SEXP processx__process_exists(SEXP pid) {
   } else if (errno == ESRCH) {
     return ScalarLogical(0);
   } else {
-    R_THROW_SYSTEM_ERROR("kill syscall error");
+    R_THROW_SYSTEM_ERROR("kill syscall error for pid '%d'", cpid);
     return R_NilValue;
   }
 }
